"""
Used to download datasets included in Scarf.

Classes:
    OSFdownloader: A class for downloading datasets from OSF.

- Methods:
    - handle_download: carry out the download of a specified dataset
    - show_available_datasets: list datasets offered through Scarf
    - fetch_dataset: Downloads datasets from online repositories and saves them in as-is format
"""

import requests
import os
import tarfile
from .logging_utils import logger
from .utils import system_call
import pandas as pd
import io

__all__ = ['show_available_datasets', 'fetch_dataset']


class OSFdownloader:
    """
        A class for downloading datasets from OSF.

        Attributes:
            projectId:
            storages:
            url:
            datasets:
            sourceFile:
            sources:

        Methods:
            get_json:
            get_all_pages:
            show_datasets:
            get_dataset_file_ids:
        """
    def __init__(self, project_id):
        self.projectId = project_id
        self.storages = ['osfstorage', 'figshare']
        self.url = f"https://api.osf.io/v2/nodes/{self.projectId}/files/"
        self.datasets, self.sourceFile = self._populate_datasets()
        self.sources = self._populate_sources()

    def get_json(self, storage, endpoint, url):
        if endpoint != '':
            endpoint = endpoint + '/'
        if url is None:
            url = self.url + f"{storage}/{endpoint}"
        return requests.get(url).json()

    def get_all_pages(self, storage, endpoint=''):
        data = []
        url = None
        while True:
            r = self.get_json(storage, endpoint, url)
            data.extend(r['data'])
            url = r['links']['next']
            if url is None:
                break
        return data

    @staticmethod
    def _process_path(node):
        return node['attributes']['path'].rstrip('/').lstrip('/')

    def _populate_datasets(self):
        datasets = {}
        source_fn = ''
        for storage in self.storages:
            for i in self.get_all_pages(storage):
                path = self._process_path(i)
                if i['attributes']['name'] == 'sources':
                    source_fn = path
                    continue
                datasets[i['attributes']['name']] = (path, storage)
        return datasets, source_fn

    def _populate_sources(self):
        source_fn = self._get_files_for_node('osfstorage', self.sourceFile)['sources.csv']
        return pd.read_csv(io.StringIO(requests.get(source_fn).text)).set_index('id').to_dict()

    def show_datasets(self):
        print('\n'.join(sorted(self.datasets.keys())))

    def _get_files_for_node(self, storage, file_id):
        base_url = f"https://files.de-1.osf.io/v1/resources/{self.projectId}/providers/"
        ret_val = {}
        for i in self.get_all_pages(storage, file_id):
            path = self._process_path(i)
            ret_val[i['attributes']['name']] = base_url + f"{storage}/{path}"
        return ret_val

    def get_dataset_file_ids(self, dataset_name):
        if dataset_name not in self.datasets:
            raise KeyError(f"ERROR: {dataset_name} was not found. "
                           f"Please choose one of the following:\n{self.show_datasets()}")
        file_id, storage = self.datasets[dataset_name]
        return self._get_files_for_node(storage, file_id)


osfd = None


def handle_download(url: str, out_fn: str) -> None:
    """
    Carry out the download of a specified dataset. Invokes appropriate command for Linux and Windows OS

    Args:
        url: URL of file to be downloaded
        out_fn: Absolute path to the file where the downloaded data is to be saved

    Returns: None

    """

    import sys

    if sys.platform == 'win32':
        cmd = 'powershell -command "(New-Object System.Net.WebClient).DownloadFile(%s, %s)"' % (
            f"'{url}'", f"'{out_fn}'")
    elif sys.platform in ['posix', 'linux']:
        cmd = f"wget --no-verbose -O {out_fn} {url}"
    else:
        raise ValueError(f"This operating system is not supported in this function. "
                         f"Please download the file manually from this URL:\n {url}\n "
                         f"Please save as: {out_fn}")
    logger.info("Download started...")
    system_call(cmd)
    logger.info(f"Download finished! File saved here: {out_fn}")


def show_available_datasets() -> None:
    """
    List datasets offered through Scarf.

    Prints the list of datasets.

    Returns:
        None
    """
    global osfd
    if osfd is None:
        osfd = OSFdownloader('zeupv')
    osfd.show_datasets()


def fetch_dataset(dataset_name: str, save_path: str = '.', as_zarr: bool = False) -> None:
    """
    Downloads datasets from online repositories and saves them in as-is format.

    Args:
        dataset_name: Name of the dataset
        save_path: Save location without name of the file
        as_zarr: If True, then a Zarr format file, if available, is downloaded instead

    Returns:
        None
    """

    zarr_ext = '.zarr.tar.gz'

    def has_zarr(entry):
        for e in entry:
            if e.endswith(zarr_ext):
                return True
        return False

    def get_zarr_entry(entry):
        for e in entry:
            if e.endswith(zarr_ext):
                return e, entry[e]
        return False, False

    global osfd
    if osfd is None:
        osfd = OSFdownloader('zeupv')

    files = osfd.get_dataset_file_ids(dataset_name)

    if as_zarr:
        if has_zarr(files) is False:
            logger.error(f"Zarr file does not exist for {dataset_name}. Nothing downloaded")
            return None

    save_dir = os.path.join(save_path, dataset_name)
    if os.path.isdir(save_dir) is False:
        os.makedirs(save_dir)

    if as_zarr:
        sp, url = get_zarr_entry(files)
        sp = os.path.abspath(os.path.join(save_dir, sp))
        handle_download(url, sp)
        logger.info(f"Extracting Zarr file for {dataset_name}")
        tar = tarfile.open(sp, "r:gz")
        tar.extractall(save_dir)
        tar.close()
    else:
        for i in files:
            if i.endswith(zarr_ext):
                continue
            sp = os.path.abspath(os.path.join(save_dir, i))
            handle_download(files[i], sp)



"""
Contains the MetaData class, which is used for storing metadata about cells and features.
"""
from zarr import hierarchy as zarr_hierarchy
from zarr import array as zarr_array
import numpy as np
import re
import pandas as pd
from typing import List, Iterable, Any, Dict, Tuple
from .feat_utils import fit_lowess
from .writers import create_zarr_obj_array
from .logging_utils import logger

__all__ = ['MetaData']


def _all_true(bools: np.ndarray) -> np.ndarray:
    """

    Args:
        bools:

    Returns:

    """
    a = bools.sum(axis=0)
    a[a < bools.shape[0]] = 0
    return a.astype(bool)


class MetaData:
    """
    MetaData class for cells and features.

    This class provides an interface to perform CRUD operations on metadata, saved in the Zarr hierarchy.
    All the changes at the metadata are synchronized on disk.

    Attributes:
        locations: The locations for where the metadata is stored.
        N: The size of the primary data.
        index: A numpy array with the indices of the cells/features.
    """

    def __init__(self, zgrp: zarr_hierarchy):
        """
        Args:
            zgrp: Zarr hierarchy object wherein metadata arrays are saved.
        """
        self.locations: Dict[str, zarr_hierarchy] = {'primary': zgrp}
        self.N = self._get_size(self.locations['primary'], strict_mode=True)
        self.index = np.array(range(self.N))

    def _get_size(self, zgrp: zarr_hierarchy, strict_mode: bool = False) -> int:
        """

        Args:
            zgrp:
            strict_mode:

        Returns:

        """
        sizes = []
        for i in zgrp.keys():
            sizes.append(zgrp[i].shape[0])
        if len(sizes) > 0:
            if len(set(sizes)) != 1:
                raise ValueError("ERROR: Metadata table is corrupted. Not all columns are of same length")
            return sizes[0]
        else:
            if strict_mode:
                raise ValueError("Attempted to get size of empty zarr group")
            else:
                return self.N

    @staticmethod
    def _col_renamer(loc: str, col: str) -> str:
        """

        Args:
            loc:
            col:

        Returns:

        """
        if loc != 'primary':
            return f"{loc}_{col}"
        return col

    def _column_map(self) -> Dict[str, Tuple[str, str]]:
        """

        Returns:

        """
        reserved_cols = ['I', 'ids', 'names']
        col_map = {x: 'primary' for x in reserved_cols}
        for loc, zgrp in self.locations.items():
            for i in zgrp.keys():
                j = self._col_renamer(loc, i)
                if j in col_map and j not in reserved_cols:
                    logger.warning(f" {i} is duplicate in metadata loc {loc}. This means something has failed "
                                   f"upstream. This is quite unexpected. Please report this issue.")
                col_map[j] = (loc, i)
        return col_map

    def _get_loc(self, column: str) -> Tuple[str, str]:
        """

        Args:
            column:

        Returns:

        """
        col_map = self._column_map()
        if column not in col_map:
            raise KeyError(f"{column} does not exist in the metadata columns.")
        loc, col = col_map[column]
        return loc, col

    def _get_array(self, column: str) -> zarr_array:
        """

        Args:
            column:

        Returns:

        """
        loc, col = self._get_loc(column)
        return self.locations[loc][col]

    def get_dtype(self, column: str) -> type:
        """
        Returns the dtype for the given column.

        Args:
            column: Column name of the table.
        """
        return self._get_array(column).dtype

    def _verify_bool(self, key: str) -> bool:
        """
        Validates if a give table column (parameter 'key') is bool type

        Args:
            key: Name of the column to query

        Returns: None

        """

        if self.get_dtype(key) != bool:
            raise TypeError("ERROR: `key` should be name of a boolean type column in Metadata table")
        return True

    def mount_location(self, zgrp: zarr_hierarchy, identifier: str) -> None:
        """

        Args:
            zgrp:
            identifier:

        Returns:

        """
        if identifier in self.locations:
            raise ValueError(f"ERROR: a location with identifier '{identifier}' already mounted")
        size = self._get_size(zgrp)
        if size != self.N:
            raise ValueError(f"ERROR: The index size of the mount location ({size}) is not same as primary ({self.N})")
        new_cols = [self._col_renamer(identifier, x) for x in zgrp.keys()]
        cols = self.columns
        conflict_names = [x for x in new_cols if x in cols]
        if len(conflict_names) > 0:
            conflict_names = ' '.join(conflict_names)
            raise ValueError(f"ERROR: These names in location conflict with existing names: {conflict_names}\n. "
                             f"Please try with a different identifier value.")
        self.locations[identifier] = zgrp

    def unmount_location(self, identifier: str) -> None:
        """

        Args:
            identifier:

        Returns:

        """
        if identifier == 'primary':
            raise ValueError("Cannot unmount the primary location")
        if identifier not in self.locations:
            logger.warning(f"{identifier} is not mounted. Nothing to unmount")
            return None
        self.locations.pop(identifier)

    @property
    def columns(self) -> List[str]:
        """

        Returns:

        """
        return list(self._column_map().keys())

    def fetch_all(self, column: str) -> np.ndarray:
        """

        Args:
            column:

        Returns:

        """
        return self._get_array(column)[:]

    def active_index(self, key: str) -> np.ndarray:
        """

        Args:
            key:

        Returns:

        """
        if self._verify_bool(key):
            return self.index[self.fetch_all(key)]
        else:
            raise ValueError("ERROR: Unexpected error when verifying boolean key. Please report this issue")

    def fetch(self, column: str, key: str = 'I') -> np.ndarray:
        """
        Get column values for only valid rows

        Args:
            column:
            key:

        Returns:

        """

        return self.fetch_all(column)[self.active_index(key)]

    def _save(self, column_name: str, values: np.ndarray, location: str = 'primary') -> None:
        """

        Args:
            column_name:
            values:

        Returns:

        """
        if location not in self.locations:
            raise KeyError(f"ERROR: '{location}' has not been mounted. Save data request failed!")
        if values.shape != (self.N,):
            raise ValueError(f"ERROR: Values are of shape: {values.shape}. Expected shape is: ({self.N},)")
        create_zarr_obj_array(self.locations[location], column_name, values, values.dtype)
        return None

    def _fill_to_index(self, values: np.array, fill_value, key: str) -> np.ndarray:
        """
        Makes sure that the array being added to the table is of the same shape.
        If the array has same shape as the table then the input array is added straightaway.
        It is assumed that the input array is in same order as table rows.

        Args:
            values:
            fill_value:
            key:

        Returns:

        """

        if len(values) == self.N:
            return values
        else:
            self._verify_bool(key)
            k = self.fetch_all(key)
            if len(values) != k.sum():
                raise ValueError(f"ERROR: `values`  are of incorrect length ({len(values)}). "
                                 f" Chosen key ({key}) has {len(k)} active rows")
            else:
                a = np.empty(self.N).astype(type(values[0]))
                a[k] = values
                a[~k] = fill_value
                return a

    def get_index_by(self, value_targets: List[Any], column: str, key: str = None) -> np.ndarray:
        """

        Args:
            value_targets:
            column:
            key:
        Returns:

        """
        if isinstance(value_targets, Iterable) and type(value_targets) != str:
            if key is None:
                values = self.fetch_all(column)
            else:
                values = self.fetch(column, key)
            value_map = {}
            for n, x in enumerate(values):
                x = x.upper()
                if x not in value_map:
                    value_map[x] = []
                value_map[x].append(n)
            ret_val = []
            missing_count = 0
            for i in value_targets:
                i = i.upper()
                if i in value_map:
                    ret_val.extend(value_map[i])
                else:
                    missing_count += 1
            if missing_count > 0:
                logger.warning(f"{missing_count} values were not found in the table column {column}")
            return np.array(ret_val)
        else:
            raise TypeError("ERROR: Please provide the `value_targets` as list")

    def index_to_bool(self, idx: np.ndarray, invert: bool = False) -> np.ndarray:
        """

        Args:
            idx:
            invert:

        Returns:

        """
        a = np.zeros(self.N, dtype=bool)
        if len(idx) > 0:
            a[idx] = True
        if invert:
            a = ~a
        return a

    def insert(self, column_name: str, values: np.array, fill_value: Any = np.NaN,
               key: str = 'I', overwrite: bool = False, location: str = 'primary', force: bool = False) -> None:
        """
        Insert a column into the table.

        Args:
            column_name (str): Name of column to modify.
            values (np.array): Values the column should contain.
            fill_value (Any = np.NaN): Value to fill unassigned slots with.
            key (str = 'I'):
            overwrite (bool = False): Should function overwrite column if it already exists?
            location (str = 'primary'):
            force (bool = False): Enforce change to column, even if column is a protected column name ('I' or 'ids').

        Returns:
            None
        """
        col = self._col_renamer(location, column_name)
        if col in ['I', 'ids'] and force is False:
            raise ValueError(f"ERROR: {col} is a protected column name in MetaData class.")
        if col in self.columns and overwrite is False:
            raise ValueError(f"ERROR: {col} already exists. Please set `overwrite` to True to overwrite.")
        if type(values) == list:
            logger.warning("'values' parameter is of `list` type and not `np.ndarray` as expected. The correct dtype "
                           "may not be assigned to the column")
            values = np.array(values)
        v = self._fill_to_index(values, fill_value, key)
        self._save(column_name, v.astype(values.dtype), location=location)
        return None

    def update_key(self, values: np.array, key) -> None:
        """
        Modify a column in the metadata table, specified with `key`.

        Args:
            values: The values to update the column with.
            key: Which column in the metadata table to update.

        Returns:
            None
        """
        v = self._fill_to_index(values, False, key)
        v = _all_true(np.array([v, self.fetch_all(key)]))
        self._save(key, v)
        return None

    def reset_key(self, key: str) -> None:
        """

        Args:
            key:

        Returns:

        """
        values = np.array([True for _ in range(self.N)]).astype(bool)
        self._save(key, values)
        return None

    def drop(self, column: str) -> None:
        """

        Args:
            column:

        Returns:

        """
        if column in ['I', 'ids', 'names']:
            raise ValueError(f"ERROR: {column} is a protected name in MetaData class. Cannot be deleted")
        # noinspection PyUnusedLocal
        loc, col = self._get_loc(column)
        del self.locations[loc][col]
        return None

    def sift(self, column: str, min_v: float = -np.Inf, max_v: float = np.Inf) -> np.ndarray:
        """

        Args:
            column:
            min_v:
            max_v:

        Returns:

        """
        values = self.fetch_all(column)
        return (values > min_v) & (values < max_v)

    def multi_sift(self, columns: List[str], lows: Iterable, highs: Iterable) -> np.ndarray:
        """

        Args:
            columns:
            lows:
            highs:

        Returns:

        """
        ret_val = _all_true(np.array([self.sift(i, j, k) for i, j, k
                                      in zip(columns, lows, highs)]))
        return ret_val

    def head(self, n: int = 5) -> pd.DataFrame:
        """

        Args:
            n:

        Returns:

        """
        df = pd.DataFrame({
            x: self.fetch_all(x)[:n] for x in self.columns
        })
        return df

    def to_pandas_dataframe(self, columns: List[str], key: str = None) -> pd.DataFrame:
        """
        Returns the requested columns as a Pandas dataframe, sorted on key.
        """
        valid_cols = self.columns
        df = pd.DataFrame({x: self.fetch_all(x) for x in columns if x in valid_cols})
        if key is not None:
            df = df.reindex(self.active_index(key))
        return df

    def grep(self, pattern: str, only_valid=False) -> List[str]:
        """

        Args:
            pattern:
            only_valid:

        Returns:

        """
        names = np.array(list(map(str.upper, self.fetch_all('names'))))
        if only_valid:
            names = names[self.active_index('I')]
        return sorted(set([x for x in names
                           if re.match(pattern.upper(), x) is not None]))

    def remove_trend(self, x: str, y: str, n_bins: int = 200,
                     lowess_frac: float = 0.1) -> np.ndarray:
        """

        Args:
            x:
            y:
            n_bins:
            lowess_frac:

        Returns:

        """
        a = fit_lowess(self.fetch(x).astype(float),
                       self.fetch(y).astype(float),
                       n_bins, lowess_frac)
        return a

    def __repr__(self):
        return f"MetaData of {self.fetch_all('I').sum()}({self.N}) elements"



"""
Utility functions for features.
"""
import pandas as pd
import numpy as np
from typing import List, Sequence

__all__ = ['fit_lowess', 'binned_sampling']


def fit_lowess(a, b, n_bins: int, lowess_frac: float) -> np.ndarray:
    """
    Fits a LOWESS (Locally Weighted Scatterplot Smoothing) curve.

    Args:
        a:
        b:
        n_bins:
        lowess_frac:

    Returns:

    """
    from statsmodels.nonparametric.smoothers_lowess import lowess

    stats = pd.DataFrame({'a': a, 'b': b}).apply(np.log)
    bin_edges = np.histogram(stats.a, bins=n_bins)[1]
    bin_edges[-1] += 0.1  # For including last gene
    bin_idx = []
    for i in range(n_bins):
        idx = pd.Series((stats.a >= bin_edges[i]) & (stats.a < bin_edges[i + 1]))
        if sum(idx) > 0:
            bin_idx.append(list(idx[idx].index))
    bin_vals = []
    for idx in bin_idx:
        temp_stat = stats.reindex(idx)
        temp_gene = temp_stat.idxmin().b
        bin_vals.append(
            [temp_stat.b[temp_gene], temp_stat.a[temp_gene]])
    bin_vals = np.array(bin_vals).T
    bin_cor_fac = lowess(bin_vals[0], bin_vals[1], return_sorted=False,
                         frac=lowess_frac, it=100).T
    fixed_var = {}
    for bcf, indices in zip(bin_cor_fac, bin_idx):
        for idx in indices:
            fixed_var[idx] = np.e ** (stats.b[idx] - bcf)
    return np.array([fixed_var[x] for x in range(len(a))])


def binned_sampling(values: pd.Series, feature_list: List[str], ctrl_size: int,
                    n_bins: int, rand_seed: int) -> List[str]:
    """
    Score a set of genes [Satija15]_.
    The score is the average expression of a set of genes subtracted with the
    average expression of a reference set of genes. The reference set is
    randomly sampled from the `gene_pool` for each binned expression value.

    This reproduces the approach in Seurat [Satija15]_ and has been implemented
    for Scanpy by Davide Cittaro.

    This function is adapted from Scanpy's `score_genes`.

    Args:
        values: The values for the features.
        feature_list: The list of features to use for score calculation.
        ctrl_size: Number of reference features to be sampled from each bin.
        n_bins: Number of bins for sampling.
        rand_seed: The seed to use for the random number generation.

    Returns:
        A list of sampled features.
    """
    n_items = int(np.round(len(values) / (n_bins - 1)))
    feature_list = set(feature_list)
    # Made following more linter friendly
    # obs_cut = obs_avg.rank(method='min') // n_items
    obs_cut: pd.Series = values.fillna(0).rank(method='min').divide(n_items).astype(int)

    control_genes = set()
    for cut in np.unique(obs_cut[feature_list]):
        # Replaced np.random.shuffle with pandas' sample method
        r_genes = obs_cut[obs_cut == cut].sample(n=ctrl_size, random_state=rand_seed).index
        control_genes.update(set(r_genes))
    return list(control_genes - feature_list)



"""
====================================
Scarf - Single-cell atlases reformed
====================================

Scarf is a Python package that performs memory-efficient analysis of single-cell genomics data.

- Analyze atlas scale datasets on your laptop (tested with up to 4 million cells)
- Perform analysis of scATAC-Seq data (datasets with up to 700K cells with 1 million peaks tested) under 10 GB RAM
- Make parallel implementations of UMAP and tSNE (SG-tSNE) for quick cell embedding
- Perform hierarchical clustering that gives interpretable cluster relationships
- Sub-sample highly representative cells using state-of-the-art method TopACeDo
- Perform quick and accurate projections of cells from one dataset onto another or integrate multiple datasets.

Exports:
--------

- Modules
    - datastore: Contains the primary interface to interact with data (i. e. DataStore) and its superclasses.
    - readers: A collection of classes for reading in different data formats.
    - writers: Methods and classes for writing data to disk.
    - meld_assay:
    - utils: Utility methods.
    - downloader: Used to download datasets included in Scarf.

GitHub: https://github.com/parashardhapola/scarf

Documentation: https://scarf.readthedocs.io/en/latest/index.html

Pre-print: https://www.biorxiv.org/content/10.1101/2021.05.02.441899v1

PyPI: https://pypi.org/project/scarf/
"""

import warnings
from dask.array import PerformanceWarning
from importlib_metadata import version

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=PerformanceWarning)

try:
    __version__ = version('scarf')
except ImportError:
    print("Scarf is not installed", flush=True)

from .datastore import *
from .readers import *
from .writers import *
from .meld_assay import *
from .utils import *
from .downloader import *




"""
A collection of classes for reading in different data formats.

- Classes:
    - CrH5Reader: A class to read in CellRanger (Cr) data, in the form of an H5 file.
    - CrDirReader: A class to read in CellRanger (Cr) data, in the form of a directory.
    - CrReader: A class to read in CellRanger (Cr) data.
    - H5adReader: A class to read in data in the form of a H5ad file (h5 file with AnnData information).
    - MtxDirReader: A class to read in data in the form of a directory containing a Matrix Market file + accompanying files.
    - NaboH5Reader: A class to read in data in the form of a Nabo H5 file.
    - LoomReader: A class to read in data in the form of a Loom file.
"""

from abc import ABC, abstractmethod
from typing import Generator, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import os
import sparse
from typing import IO
import h5py
from tqdm import tqdm
from .logging_utils import logger

__all__ = ['CrH5Reader', 'CrDirReader', 'CrReader', 'H5adReader', 'MtxDirReader', 'NaboH5Reader', 'LoomReader']


def get_file_handle(fn: str) -> IO:
    """
    Returns a file object for the given file name.

    Args:
        fn: The path to the file (file name).
    """
    import gzip

    try:
        if fn.rsplit('.', 1)[-1] == 'gz':
            return gzip.open(fn, mode='rt')
        else:
            return open(fn, 'r')
    except (OSError, IOError, FileNotFoundError):
        raise FileNotFoundError("ERROR: FILE NOT FOUND: %s" % fn)


def read_file(fn: str):
    """
    Yields the lines from the file the given file name points to.

    Args:
        fn: The path to the file (file name).
    """
    fh = get_file_handle(fn)
    for line in fh:
        yield line.rstrip()


class CrReader(ABC):
    """
    A class to read in CellRanger (Cr) data.

    Attributes:
        autoNames: Specifies if the data is from RNA or ATAC sequencing.
        grpNames: A dictionary that specifies where to find the matrix, features and barcodes.
        nFeatures: Number of features in dataset.
        nCells: Number of cells in dataset.
        assayFeats: A DataFrame with information about the features in the assay.
    """
    def __init__(self, grp_names, file_type):
        """
        Args:
            grp_names (Dict): A dictionary that specifies where to find the matrix, features and barcodes.
            file_type (str): Type of sequencing data ('rna' | 'atac')
        """
        if file_type == 'rna':
            self.autoNames = {'Gene Expression': 'RNA'}
        elif file_type == 'atac':
            self.autoNames = {'Peaks': 'ATAC'}
        else:
            raise ValueError("ERROR: Please provide a value for parameter 'file_type'.\n"
                             "The value can be either 'rna' or 'atac' depending on whether is is scRNA-Seq or "
                             "scATAC-Seq data")
        self.grpNames: Dict = grp_names
        self.nFeatures: int = len(self.feature_names())
        self.nCells: int = len(self.cell_names())
        self.assayFeats = self._make_feat_table()
        self._auto_rename_assay_names()

    @abstractmethod
    def _handle_version(self):
        pass

    @abstractmethod
    def _read_dataset(self, key: Optional[str] = None) -> List:
        pass

    @abstractmethod
    def consume(self, batch_size: int, lines_in_mem: int):
        """
        Returns a generator that yield chunks of data.
        """
        pass

    def _subset_by_assay(self, v, assay) -> List:
        if assay is None:
            return v
        elif assay not in self.assayFeats:
            raise ValueError("ERROR: Assay ID %s is not valid" % assay)
        idx = self.assayFeats[assay]
        return v[idx.start: idx.end]

    def _make_feat_table(self) -> pd.DataFrame:
        s = self.feature_types()
        span: List[Tuple] = []
        last = s[0]
        last_n: int = 0
        for n, i in enumerate(s[1:], 1):
            if i != last:
                span.append((last, last_n, n))
                last_n = n
            elif n == len(s) - 1:
                span.append((last, last_n, n + 1))
            last = i
        df = pd.DataFrame(span, columns=['type', 'start', 'end'])
        df.index = ["assay%s" % str(x + 1) for x in df.index]
        df['nFeatures'] = df.end - df.start
        return df.T

    def _auto_rename_assay_names(self):
        anames = list(map(str.upper, self.assayFeats.columns))
        main_name_k = list(self.autoNames.keys())[0]
        main_name_v = list(self.autoNames.values())[0]
        if main_name_v in anames:
            logger.info(f"{main_name_v} already present")
            # Making sure that column name is in uppercase 'RNA'
            newnames = list(self.assayFeats.columns)
            newnames[anames.index(main_name_v)] = main_name_v
            self.assayFeats.columns = newnames
        else:
            at = self.assayFeats.T['type'] == main_name_k
            if at.sum() == 1:
                main_assay = at[at].index[0]
            else:
                # FIXME: raise warning here
                main_assay = self.assayFeats.T[at].nFeatures.astype(int).idxmax()
            self.rename_assays({main_assay: main_name_v})

    def rename_assays(self, name_map: Dict[str, str]) -> None:
        """
        Renames specified assays in the Reader.

        Args:
            name_map: A Dictionary containing current name as key and new name as value.
        """
        self.assayFeats.rename(columns=name_map, inplace=True)

    def feature_ids(self, assay: str = None) -> List[str]:
        """
        Returns a list of feature IDs in a specified assay.

        Args:
            assay: Select which assay to retrieve feature IDs from.
        """
        return self._subset_by_assay(
            self._read_dataset('feature_ids'), assay)

    def feature_names(self, assay: str = None) -> List[str]:
        """
        Returns a list of features in the dataset.

        Args:
            assay: Select which assay to retrieve features from.
        """
        vals = self._read_dataset('feature_names')
        if vals is None:
            logger.warning('Feature names extraction failed using feature IDs')
            vals = self._read_dataset('feature_ids')
        return self._subset_by_assay(vals, assay)

    def feature_types(self) -> List[str]:
        """
        Returns a list of feature types in the dataset.
        """
        if self.grpNames['feature_types'] is not None:
            ret_val = self._read_dataset('feature_types')
            if ret_val is not None:
                return ret_val
        default_name = list(self.autoNames.keys())[0]
        return [default_name for _ in range(self.nFeatures)]

    def cell_names(self) -> List[str]:
        """
        Returns a list of names of the cells in the dataset.
        """
        return self._read_dataset('cell_names')


class CrH5Reader(CrReader):
    """
    A class to read in CellRanger (Cr) data, in the form of an H5 file.

    Subclass of CrReader.

    Attributes:
        autoNames: Specifies if the data is from RNA or ATAC sequencing.
        grpNames: A dictionary that specifies where to find the matrix, features and barcodes.
        nFeatures: Number of features in dataset.
        nCells: Number of cells in dataset.
        assayFeats: A DataFrame with information about the features in the assay.
        h5obj: A File object from the h5py package.
        grp: Current active group in the hierarchy.
    """
    def __init__(self, h5_fn, file_type: str = None):
        """
        Args:
            h5_fn: File name for the h5 file.
            file_type (str): Type of sequencing data ('rna' | 'atac')
        """
        self.h5obj = h5py.File(h5_fn, mode='r')
        self.grp = None
        super().__init__(self._handle_version(), file_type)

    def _handle_version(self):
        root_key = list(self.h5obj.keys())[0]
        self.grp = self.h5obj[root_key]
        if root_key == 'matrix':
            grps = {'feature_ids': 'features/id',
                    'feature_names': 'features/name',
                    'feature_types': 'features/feature_type',
                    'cell_names': 'barcodes'}
        else:
            grps = {'feature_ids': 'genes', 'feature_names': 'gene_names',
                    'feature_types': None, 'cell_names': 'barcodes'}
        return grps

    def _read_dataset(self, key: Optional[str] = None):
        return [x.decode('UTF-8') for x in self.grp[self.grpNames[key]][:]]

    # noinspection DuplicatedCode
    def consume(self, batch_size: int, lines_in_mem: int):
        s = 0
        for ind_n in range(0, self.nCells, batch_size):
            i = self.grp['indptr'][ind_n:ind_n + batch_size]
            e = i[-1]
            if s != 0:
                idx = np.array([s] + list(i))
                idx = idx - idx[0]
            else:
                idx = np.array(i)
            n = idx.shape[0] - 1
            nidx = np.repeat(range(n), np.diff(idx).astype('int32'))
            yield sparse.COO([nidx, self.grp['indices'][s: e]], self.grp['data'][s: e],
                             shape=(n, self.nFeatures))
            s = e

    def close(self) -> None:
        """
        Closes file connection.
        """
        self.h5obj.close()


class CrDirReader(CrReader):
    """
    A class to read in CellRanger (Cr) data, in the form of a directory.

    Subclass of CrReader.

    Attributes:
        loc: Path for the directory containing the cellranger output.
        matFn: The file name for the matrix file.
    """
    def __init__(self, loc, file_type: str = None):
        """
        Args:
            loc (str): Path for the directory containing the cellranger output.
            file_type (str): Type of sequencing data ('rna' | 'atac')
        """
        self.loc: str = loc.rstrip('/') + '/'
        self.matFn = None
        super().__init__(self._handle_version(), file_type)

    def _handle_version(self):
        if os.path.isfile(self.loc + 'features.tsv.gz'):
            self.matFn = self.loc + 'matrix.mtx.gz'
            grps = {'feature_ids': ('features.tsv.gz', 0),
                    'feature_names': ('features.tsv.gz', 1),
                    'feature_types': ('features.tsv.gz', 2),
                    'cell_names': ('barcodes.tsv.gz', 0)}
        elif os.path.isfile(self.loc + 'genes.tsv'):
            self.matFn = self.loc + 'matrix.mtx'
            grps = {'feature_ids': ('genes.tsv', 0),
                    'feature_names': ('genes.tsv', 1),
                    'feature_types': None,
                    'cell_names': ('barcodes.tsv', 0)}
        else:
            raise IOError("ERROR: Couldn't find either of these expected combinations of files:\n"
                          "\t- matrix.mtx, barcodes.tsv and genes.tsv\n"
                          "\t- matrix.mtx.gz, barcodes.tsv.gz and features.tsv.gz\n"
                          "Please make sure that you have not compressed or uncompressed the Cellranger output files "
                          "manually")
        return grps

    def _read_dataset(self, key: Optional[str] = None):
        try:
            vals = [x.split('\t')[self.grpNames[key][1]] for x in
                    read_file(self.loc + self.grpNames[key][0])]
        except IndexError:
            logger.warning(f"{key} extraction failed from {self.grpNames[key][0]} "
                           f"in column {self.grpNames[key][1]}", flush=True)
            vals = None
        return vals

    # noinspection DuplicatedCode
    def to_sparse(self, a: np.ndarray) -> sparse.COO:
        """
        Returns the input data as a sparse (COO) matrix.

        Args:
            a: Sparse matrix, contains a chunk of data from the MTX file.
        """
        idx = np.where(np.diff(a[:, 1]) == 1)[0] + 1
        return sparse.COO([(a[:, 1] - a[0, 1]).astype(int),
                           (a[:, 0] - 1).astype(int)],
                          a[:, 2], shape=(len(idx) + 1, self.nFeatures))

    # noinspection DuplicatedCode
    def consume(self, batch_size: int, lines_in_mem: int = int(1e5)) -> \
            Generator[List[np.ndarray], None, None]:
        stream = pd.read_csv(self.matFn, skiprows=3, sep=' ',
                             header=None, chunksize=lines_in_mem)
        start = 1
        dfs = []
        for df in stream:
            if df.iloc[-1, 1] - start >= batch_size:
                idx = df[1] < batch_size + start
                dfs.append(df[idx])
                yield self.to_sparse(np.vstack(dfs))
                dfs = [df[~idx]]
                start += batch_size
            else:
                dfs.append(df)
        yield self.to_sparse(np.vstack(dfs))


class MtxDirReader(CrReader):
    """
    A class to read a directory with a Matrix Market file and its accompanying files.

    Subclass of CrReader.

    Attributes:
        loc: Path for the directory containing the cellranger output.
        matFn: The file name for the matrix file.
    """
    def __init__(self, loc, file_type: str = None):
        """
        Args:
            loc (str): Path for the directory containing the cellranger output.
            file_type (str): Type of sequencing data ('rna' | 'atac')
        """
        self.loc: str = loc.rstrip('/') + '/'
        self.matFn = None
        super().__init__(self._handle_version(), file_type)

    def _handle_version(self):
        if os.path.isfile(self.loc + 'features.tsv.gz'):
            self.matFn = self.loc + 'matrix.mtx.gz'
            grps = {'feature_ids': ('features.tsv.gz', 0),
                    'feature_names': ('features.tsv.gz', 1),
                    'feature_types': ('features.tsv.gz', 2),
                    'cell_names': ('barcodes.tsv.gz', 0)}
        elif os.path.isfile(self.loc + 'features.tsv.gz'):  # sometimes somebody might have gunziped these files...
            self.matFn = self.loc + 'matrix.mtx'
            grps = {'feature_ids': ('features.tsv', 0),
                    'feature_names': ('features.tsv', 1),
                    'feature_types': None,
                    'cell_names': ('barcodes.tsv', 0)}
        elif os.path.isfile(self.loc + 'genes.tsv'):
            self.matFn = self.loc + 'matrix.mtx'
            grps = {'feature_ids': ('genes.tsv', 0),
                    'feature_names': ('genes.tsv', 1),
                    'feature_types': None,
                    'cell_names': ('barcodes.tsv', 0)}
        elif os.path.isfile(self.loc + 'genes.tsv.gz'):
            self.matFn = self.loc + 'matrix.mtx.gz'
            grps = {'feature_ids': ('genes.tsv.gz', 0),
                    'feature_names': ('genes.tsv.gz', 1),
                    'feature_types': ('genes.tsv.gz', 2),
                    'cell_names': ('barcodes.tsv.gz', 0)}
        else:
            raise IOError("ERROR: Couldn't find either of these expected combinations of files:\n"
                          "\t- matrix.mtx, barcodes.tsv and genes.tsv\n"
                          "\t- matrix.mtx.gz, barcodes.tsv.gz and features.tsv.gz\n"
                          "Please make sure that you have not compressed or uncompressed the Cellranger output files "
                          "manually")
        return grps

    def _read_dataset(self, key: Optional[str] = None):
        try:
            vals = [x.split('\t')[self.grpNames[key][1]] for x in
                    read_file(self.loc + self.grpNames[key][0])]
        except IndexError:
            vals = None
        return vals

    # noinspection DuplicatedCode
    def to_sparse(self, a):
        """
        Returns the input data as a sparse (COO) matrix.

        Args:
            a: a dense numpy matrix
        """
        idx = np.where(np.diff(a[:, 1]) == 1)[0] + 1
        return sparse.COO([(a[:, 1] - a[0, 1]).astype(int), (a[:, 0] - 1).astype(int)],
                          a[:, 2], shape=(len(idx) + 1, self.nFeatures))

    def _subset_by_assay(self, v, assay) -> List:
        if assay is None:
            return v
        elif assay not in self.assayFeats:
            raise ValueError("ERROR: Assay ID %s is not valid" % assay)
        if len(self.assayFeats[assay].shape) == 2:
            ret_val = []
            for i in self.assayFeats[assay].values[1:3].T:
                ret_val.extend(list(v[i[0]: i[1]]))
            return ret_val
        elif len(self.assayFeats[assay].shape) == 1:
            idx = self.assayFeats[assay]
            return v[idx.start: idx.end]
        else:
            raise ValueError("ERROR: assay feats is 3D. Something went really wrong. Create a github issue")

    # noinspection DuplicatedCode
    def consume(self, batch_size: int, lines_in_mem: int = int(1e5)) -> \
            Generator[List[np.ndarray], None, None]:
        stream = pd.read_csv(self.matFn, skiprows=3, sep='\t',
                             header=None, chunksize=lines_in_mem)
        start = 1
        dfs = []
        for df in stream:
            if df.iloc[-1, 1] - start >= batch_size:
                idx = df[1] < batch_size + start
                dfs.append(df[idx])
                yield self.to_sparse(np.vstack(dfs))
                dfs = [df[~idx]]
                start += batch_size
            else:
                dfs.append(df)
        yield self.to_sparse(np.vstack(dfs))


class H5adReader:
    """
    A class to read in data from a H5ad file (h5 file with AnnData information).

    Attributes:
        h5: A File object from the h5py package.
        matrix_key: Group where in the sparse matrix resides (default: 'X')
        cellAttrsKey: Group wherein the cell attributes are present
        featureAttrsKey: Group wherein the feature attributes are present
        groupCodes: Used to ensure compatibility with different AnnData versions.
        nFeatures: Number of features in dataset.
        nCells: Number of cells in dataset.
        cellIdsKey: Key in `obs` group that contains unique cell IDs. By default the index will be used.
        featIdsKey: Key in `var` group that contains unique feature IDs. By default the index will be used.
        featNamesKey: Key in `var` group that contains feature names. (Default: gene_short_name)
        catNamesKey: Looks up this group and replaces the values in `var` and 'obs' child datasets with the
                     corresponding index value within this group.
        matrixDtype: dtype of the matrix containing the data (as indicated by matrix_key)
    """
    def __init__(self, h5ad_fn: str, cell_attrs_key: str = 'obs', cell_ids_key: str = '_index',
                 feature_attrs_key: str = 'var', feature_ids_key: str = '_index',
                 feature_name_key: str = 'gene_short_name',  matrix_key: str = 'X',
                 category_names_key: str = '__categories', dtype: str = None):
        """
        Args:
            h5ad_fn: Path to H5AD file
            cell_attrs_key: H5 group under which cell attributes are saved.(Default value: 'obs')
            feature_attrs_key: H5 group under which feature attributes are saved.(Default value: 'var')
            cell_ids_key: Key in `obs` group that contains unique cell IDs. By default the index will be used.
            feature_ids_key: Key in `var` group that contains unique feature IDs. By default the index will be used.
            feature_name_key: Key in `var` group that contains feature names. (Default: gene_short_name)
            matrix_key: Group where in the sparse matrix resides (default: 'X')
            category_names_key: Looks up this group and replaces the values in `var` and 'obs' child datasets with the
                                corresponding index value within this group.
            dtype: Numpy dtype of the matrix data. This dtype is enforced when streaming the data through `consume`
                   method. (Default value: Automatically determined)
        """

        self.h5 = h5py.File(h5ad_fn, mode='r')
        self.matrixKey = matrix_key
        self.cellAttrsKey, self.featureAttrsKey = cell_attrs_key, feature_attrs_key
        self.groupCodes = {self.cellAttrsKey: self._validate_group(self.cellAttrsKey),
                           self.featureAttrsKey: self._validate_group(self.featureAttrsKey),
                           self.matrixKey: self._validate_group(self.matrixKey)}
        self.nCells, self.nFeatures = self._get_n(self.cellAttrsKey), self._get_n(self.featureAttrsKey)
        self.cellIdsKey = self._fix_name_key(self.cellAttrsKey, cell_ids_key)
        self.featIdsKey = self._fix_name_key(self.featureAttrsKey, feature_ids_key)
        self.featNamesKey = feature_name_key
        self.catNamesKey = category_names_key
        self.matrixDtype = self._get_matrix_dtype() if dtype is None else dtype

    def _validate_group(self, group: str) -> int:
        if group not in self.h5:
            logger.warning(f"`{group}` group not found in the H5ad file")
            ret_val = 0
        elif type(self.h5[group]) == h5py.Dataset:
            ret_val = 1
        elif type(self.h5[group]) == h5py.Group:
            ret_val = 2
        else:
            logger.warning(f"`{group}` slot in H5ad file is not of Dataset or Group type. "
                           f"Due to this, no information in `{group}` can be used")
            ret_val = 0
        if ret_val == 2:
            if len(self.h5[group].keys()) == 0:
                logger.warning(f"`{group}` slot in H5ad file is empty.")
                ret_val = 0
            elif len(set([self.h5[group][x].shape[0] for x in self.h5[group].keys() if
                          type(self.h5[group][x]) == h5py.Dataset])) > 1:
                if sorted(self.h5[group].keys()) != ['data', 'indices', 'indptr']:
                    logger.info(f"`{group}` slot in H5ad file has unequal sized child groups")
        return ret_val

    def _get_matrix_dtype(self):
        if self.groupCodes[self.matrixKey] == 1:
            return self.h5[self.matrixKey].dtype
        elif self.groupCodes[self.matrixKey] == 2:
            return self.h5[self.matrixKey]['data'].dtype
        else:
            raise ValueError(f"ERROR: {self.matrixKey} is neither Dataset or Group type. Will not consume data")

    def _check_exists(self, group: str, key: str) -> bool:
        if group in self.groupCodes:
            group_code = self.groupCodes[group]
        else:
            group_code = self._validate_group(group)
            self.groupCodes[group] = group_code
        if group_code == 1:
            if key in list(self.h5[group].dtype.names):
                return True
        if group_code == 2:
            if key in self.h5[group].keys():
                return True
        return False

    def _fix_name_key(self, group: str, key: str) -> str:
        if self._check_exists(group, key) is False:
            if key.startswith('_'):
                temp_key = key[1:]
                if self._check_exists(group, temp_key):
                    return temp_key
        return key

    def _get_n(self, group: str) -> int:
        if self.groupCodes[group] == 0:
            if self._check_exists(self.matrixKey, 'shape'):
                return self.h5[self.matrixKey]['shape'][0]
            else:
                raise KeyError(f"ERROR: `{group}` not found and `shape` key is missing in the {self.matrixKey} group. "
                               f"Aborting read process.")
        elif self.groupCodes[group] == 1:
            return self.h5[group].shape[0]
        else:
            for i in self.h5[group].keys():
                if type(self.h5[group][i]) == h5py.Dataset:
                    return self.h5[group][i].shape[0]
            raise KeyError(f"ERROR: `{group}` key doesn't contain any child node of Dataset type."
                           f"Aborting because unexpected H5ad format.")

    def cell_ids(self) -> np.ndarray:
        """
        Returns a list of cell IDs.
        """
        if self._check_exists(self.cellAttrsKey, self.cellIdsKey):
            if self.groupCodes[self.cellAttrsKey] == 1:
                return self.h5[self.cellAttrsKey][self.cellIdsKey]
            else:
                return self.h5[self.cellAttrsKey][self.cellIdsKey][:]
        logger.warning(f"Could not find cells ids key: {self.cellIdsKey} in `obs`.")
        return np.array([f'cell_{x}' for x in range(self.nCells)])

    # noinspection DuplicatedCode
    def feat_ids(self) -> np.ndarray:
        """
        Returns a list of feature IDs.
        """
        if self._check_exists(self.featureAttrsKey, self.featIdsKey):
            if self.groupCodes[self.featureAttrsKey] == 1:
                return self.h5[self.featureAttrsKey][self.featIdsKey]
            else:
                return self.h5[self.featureAttrsKey][self.featIdsKey][:]
        logger.warning(f"Could not find feature ids key: {self.featIdsKey} in {self.featureAttrsKey}.")
        return np.array([f'feature_{x}' for x in range(self.nFeatures)])

    # noinspection DuplicatedCode
    def feat_names(self) -> np.ndarray:
        """
        Returns a list of feature names.
        """
        if self._check_exists(self.featureAttrsKey, self.featNamesKey):
            if self.groupCodes[self.featureAttrsKey] == 1:
                values = self.h5[self.featureAttrsKey][self.featNamesKey]
            else:
                values = self.h5[self.featureAttrsKey][self.featNamesKey][:]
            return self._replace_category_values(values, self.featNamesKey, self.featureAttrsKey).astype(object)
        logger.warning(f"Could not find feature names key: {self.featNamesKey} in self.featureAttrsKey.")
        return self.feat_ids()

    def _replace_category_values(self, v: np.ndarray, key: str, group: str):
        if self.catNamesKey is not None:
            if self._check_exists(group, self.catNamesKey):
                cat_g = self.h5[group][self.catNamesKey]
                if type(cat_g) == h5py.Group:
                    if key in cat_g:
                        c = cat_g[key][:]
                        try:
                            return np.array([c[x] for x in v])
                        except (IndexError, TypeError):
                            return v
        if 'uns' in self.h5:
            if key+'_categories' in self.h5['uns']:
                c = self.h5['uns'][key+'_categories'][:]
                try:
                    return np.array([c[x] for x in v])
                except (IndexError, TypeError):
                    return v
        return v

    def _get_col_data(self, group: str, ignore_keys: List[str]) -> Generator[Tuple[str, np.ndarray], None, None]:
        if self.groupCodes[group] == 1:
            for i in tqdm(self.h5[group].dtype.names, desc=f"Reading attributes from group {group}"):
                if i in ignore_keys:
                    continue
                yield i, self._replace_category_values(self.h5[group][i][:], i, group)
        if self.groupCodes[group] == 2:
            for i in tqdm(self.h5[group].keys(), desc=f"Reading attributes from group {group}"):
                if i in ignore_keys:
                    continue
                if type(self.h5[group][i]) == h5py.Dataset:
                    yield i, self._replace_category_values(self.h5[group][i][:], i, group)

    def get_cell_columns(self) -> Generator[Tuple[str, np.ndarray], None, None]:
        """
        Creates a Generator that yields the cell columns.
        """
        for i, j in self._get_col_data(self.cellAttrsKey, [self.cellIdsKey]):
            yield i, j

    def get_feat_columns(self) -> Generator[Tuple[str, np.ndarray], None, None]:
        """
        Creates a Generator that yields the feature columns.
        """
        for i, j in self._get_col_data(self.featureAttrsKey, [self.featIdsKey, self.featNamesKey]):
            yield i, j

    # noinspection DuplicatedCode
    def consume_dataset(self, batch_size: int = 1000) -> Generator[sparse.COO, None, None]:
        """
        Returns a generator that yield chunks of data.
        """
        dset = self.h5[self.matrixKey]
        s = 0
        for e in range(batch_size, dset.shape[0] + batch_size, batch_size):
            if e > dset.shape[0]:
                e = dset.shape[0]
            yield dset[s:e]
            s = e

    def consume_group(self, batch_size: int) -> Generator[sparse.COO, None, None]:
        """
        Returns a generator that yield chunks of data.
        """
        grp = self.h5[self.matrixKey]
        s = 0
        for ind_n in range(0, self.nCells, batch_size):
            i = grp['indptr'][ind_n:ind_n + batch_size]
            e = i[-1]
            if s != 0:
                idx = np.array([s] + list(i))
                idx = idx - idx[0]
            else:
                idx = np.array(i)
            n = idx.shape[0] - 1
            nidx = np.repeat(range(n), np.diff(idx).astype('int32'))
            yield sparse.COO([nidx, grp['indices'][s: e]], grp['data'][s: e],
                             shape=(n, self.nFeatures)).todense()
            s = e

    def consume(self, batch_size: int = 1000):
        """
        Returns a generator that yield chunks of data.
        """
        if self.groupCodes[self.matrixKey] == 1:
            return self.consume_dataset(batch_size)
        elif self.groupCodes[self.matrixKey] == 2:
            return self.consume_group(batch_size)


class NaboH5Reader:
    """
    A class to read in data in the form of a Nabo H5 file.

    Attributes:
        h5: A File object from the h5py package.
        nCells: Number of cells in dataset.
        nFeatures: Number of features in dataset.
    """
    def __init__(self, h5_fn: str):
        """
        Args:
            h5_fn: Path to H5 file.
        """
        self.h5 = h5py.File(h5_fn, mode='r')
        self._check_integrity()
        self.nCells = self.h5['names']['cells'].shape[0]
        self.nFeatures = self.h5['names']['genes'].shape[0]

    def _check_integrity(self) -> bool:
        for i in ['cell_data', 'gene_data', 'names']:
            if i not in self.h5:
                raise KeyError(f"ERROR: Expected group: {i} is missing in the H5 file")
        return True

    def cell_ids(self) -> List[str]:
        """
        Returns a list of cell IDs.
        """
        return [x.decode('UTF-8') for x in self.h5['names']['cells'][:]]

    def feat_ids(self) -> np.ndarray:
        """
        Returns a list of feature IDs.
        """
        return np.array([f'feature_{x}' for x in range(self.nFeatures)])

    def feat_names(self) -> List[str]:
        """
        Returns a list of feature names.
        """
        return [x.decode('UTF-8').rsplit('_', 1)[0] for x in self.h5['names']['genes'][:]]

    def consume(self, batch_size: int = 100) -> Generator[np.ndarray, None, None]:
        """
        Returns a generator that yield chunks of data.
        """
        batch = []
        for i in self.h5['cell_data']:
            a = np.zeros(self.nFeatures).astype(int)
            v = self.h5['cell_data'][i][:][::-1]
            a[v['idx']] = v['val']
            batch.append(a)
            if len(batch) >= batch_size:
                batch = np.array(batch)
                yield batch
                batch = []
        if len(batch) > 0:
            yield np.array(batch)


class LoomReader:
    """
    A class to read in data in the form of a Loom file.

    Attributes:
        h5: A File object from the h5py package.
        matrixKey: Child node under HDF5 file root wherein the chunked matrix is stored.
        cellAttrsKey: Child node under the HDF5 file wherein the cell attributes are stored.
        featureAttrsKey: Child node under the HDF5 file wherein the feature/gene attributes are stored.
        cellNamesKey: Child node under the `cell_attrs_key` wherein the cell names are stored.
        featureNamesKey: Child node under the `feature_attrs_key` wherein the feature/gene names are stored.
        featureIdsKey: Child node under the `feature_attrs_key` wherein the feature/gene ids are stored.
        matrixDtype: Numpy dtype of the matrix data.
        nFeatures: Number of features in dataset.
        nCells: Number of cells in dataset.
    """
    def __init__(self, loom_fn: str, matrix_key: str = 'matrix',
                 cell_attrs_key='col_attrs', cell_names_key: str = 'obs_names',
                 feature_attrs_key: str = 'row_attrs', feature_names_key: str = 'var_names',
                 feature_ids_key: str = None, dtype: str = None) -> None:
        """
        Args:
            loom_fn: Path to loom format file.
            matrix_key: Child node under HDF5 file root wherein the chunked matrix is stored. (Default value: matrix).
                        This matrix is expected to be of form (nFeatures x nCells)
            cell_attrs_key: Child node under the HDF5 file wherein the cell attributes are stored.
                            (Default value: col_attrs)
            cell_names_key: Child node under the `cell_attrs_key` wherein the cell names are stored.
                            (Default value: obs_names)
            feature_attrs_key: Child node under the HDF5 file wherein the feature/gene attributes are stored.
                               (Default value: row_attrs)
            feature_names_key: Child node under the `feature_attrs_key` wherein the feature/gene names are stored.
                               (Default value: var_names)
            feature_ids_key: Child node under the `feature_attrs_key` wherein the feature/gene ids are stored.
                               (Default value: None)
            dtype: Numpy dtype of the matrix data. This dtype is enforced when streaming the data through `consume`
                   method. (Default value: Automatically determined)
        """
        self.h5 = h5py.File(loom_fn, mode='r')
        self.matrixKey = matrix_key
        self.cellAttrsKey, self.featureAttrsKey = cell_attrs_key, feature_attrs_key
        self.cellNamesKey, self.featureNamesKey = cell_names_key, feature_names_key
        self.featureIdsKey = feature_ids_key
        self.matrixDtype = self.h5[self.matrixKey].dtype if dtype is None else dtype
        self._check_integrity()
        self.nFeatures, self.nCells = self.h5[self.matrixKey].shape

    def _check_integrity(self) -> bool:
        if self.matrixKey not in self.h5:
            raise KeyError(f"ERROR: Matrix key (location): {self.matrixKey} is missing in the H5 file")
        if self.cellAttrsKey not in self.h5:
            logger.warning(f"Cell attributes are missing. Key {self.cellAttrsKey} was not found")
        if self.featureAttrsKey not in self.h5:
            logger.warning(f"Feature attributes are missing. Key {self.featureAttrsKey} was not found")
        return True

    def cell_names(self) -> List[str]:
        """
        Returns a list of names of the cells in the dataset.
        """
        if self.cellAttrsKey not in self.h5:
            pass
        elif self.cellNamesKey not in self.h5[self.cellAttrsKey]:
            logger.warning(f"Cell names/ids key ({self.cellNamesKey}) is missing in attributes")
        else:
            return self.h5[self.cellAttrsKey][self.cellNamesKey][:]
        return [f"cell_{x}" for x in range(self.nCells)]

    def cell_ids(self) -> List[str]:
        """
        Returns a list of cell IDs.
        """
        return self.cell_names()

    def _stream_attrs(self, key, ignore) -> Generator[Tuple[str, np.ndarray], None, None]:
        if key in self.h5:
            for i in tqdm(self.h5[key].keys(), desc=f"Reading {key} attributes"):
                if i in [ignore]:
                    continue
                vals = self.h5[key][i][:]
                if vals.dtype.names is None:
                    yield i, vals
                else:
                    # Attribute is a structured array
                    for j in vals.dtype.names:
                        yield i + '_' + str(j), vals[j]

    def get_cell_attrs(self) -> Generator[Tuple[str, np.ndarray], None, None]:
        """
        Returns a Generator that yields the cells' attributes.
        """
        return self._stream_attrs(self.cellAttrsKey, [self.cellNamesKey])

    def feature_names(self) -> List[str]:
        """
        Returns a list of feature names.
        """
        if self.featureAttrsKey not in self.h5:
            pass
        elif self.featureNamesKey not in self.h5[self.featureAttrsKey]:
            logger.warning(f"Feature names key ({self.featureNamesKey}) is missing in attributes")
        else:
            return self.h5[self.featureAttrsKey][self.featureNamesKey][:]
        return [f"feature_{x}" for x in range(self.nFeatures)]

    def feature_ids(self) -> List[str]:
        """
        Returns a list of feature IDs.
        """
        if self.featureAttrsKey not in self.h5:
            pass
        elif self.featureIdsKey is None:
            pass
        elif self.featureIdsKey not in self.h5[self.featureAttrsKey]:
            logger.warning(f"Feature names key ({self.featureIdsKey}) is missing in attributes")
        else:
            return self.h5[self.featureAttrsKey][self.featureIdsKey][:]
        return [f"feature_{x}" for x in range(self.nFeatures)]

    def get_feature_attrs(self) -> Generator[Tuple[str, np.ndarray], None, None]:
        """
        Returns a Generator that yields the features' attributes.
        """
        return self._stream_attrs(self.featureAttrsKey,
                                  [self.featureIdsKey, self.featureNamesKey])

    def consume(self, batch_size: int = 1000) -> Generator[np.ndarray, None, None]:
        """
        Returns a generator that yield chunks of data.
        """
        last_n = 0
        for i in range(batch_size, self.nCells, batch_size):
            yield self.h5[self.matrixKey][:, last_n:i].astype(self.matrixDtype).T
            last_n = i
        if last_n < self.nCells:
            yield self.h5[self.matrixKey][:, last_n:].astype(self.matrixDtype).T



"""
Utility functions for the mapping.
"""
import dask.array as daskarr
import numpy as np
from typing import Tuple
from .assay import Assay
from .utils import controlled_compute, show_progress
from .logging_utils import logger
import pandas as pd

__all__ = ['align_features', 'coral']


def _cov_diaged(da: daskarr) -> daskarr:
    a = daskarr.cov(da, rowvar=0)
    a[a == np.inf] = 0
    a[a == np.nan] = 0
    return a + np.eye(a.shape[0])


def _correlation_alignment(s: daskarr, t: daskarr, nthreads: int) -> daskarr:
    from scipy.linalg import fractional_matrix_power as fmp
    from threadpoolctl import threadpool_limits

    s_cov = show_progress(_cov_diaged(s), f"CORAL: Computing source covariance", nthreads)
    t_cov = show_progress(_cov_diaged(t), f"CORAL: Computing target covariance", nthreads)
    logger.info("Calculating fractional power of covariance matrices. This might take a while... ")
    with threadpool_limits(limits=nthreads):
        a_coral = np.dot(fmp(s_cov, -0.5), fmp(t_cov, 0.5))
    logger.info("Fractional power calculation complete")
    return daskarr.dot(s, a_coral)


def coral(source_data, target_data, assay, feat_key: str, nthreads: int):
    """
    Applies CORAL error correction to the input data.

    Args:
        source_data ():
        target_data ():
        assay ():
        feat_key ():
        nthreads ():
    """
    from .writers import dask_to_zarr
    from .utils import clean_array

    sm = clean_array(show_progress(source_data.mean(axis=0),
                                   'INFO: (CORAL) Calculating source feature means', nthreads))
    sd = clean_array(show_progress(source_data.std(axis=0),
                                   'INFO: (CORAL) Calculating source feature stdev', nthreads), 1)
    tm = clean_array(show_progress(target_data.mean(axis=0),
                                   'INFO: (CORAL) Calculating target feature means', nthreads))
    td = clean_array(show_progress(target_data.std(axis=0),
                                   'INFO: (CORAL) Calculating target feature stdev', nthreads), 1)
    data = _correlation_alignment((source_data - sm) / sd, (target_data - tm) / td, nthreads)
    dask_to_zarr(data, assay.z['/'], f"{assay.name}/normed__I__{feat_key}/data_coral", 1000, nthreads,
                 msg="Writing out coral corrected data")


def _order_features(s_assay, t_assay, s_feat_ids: np.ndarray, filter_null: bool,
                    exclude_missing: bool, nthreads: int) -> Tuple[np.ndarray, np.ndarray]:
    s_ids = pd.Series(s_assay.feats.fetch_all('ids'))
    t_ids = pd.Series(t_assay.feats.fetch_all('ids'))
    t_idx = t_ids.isin(s_feat_ids)
    if t_idx.sum() == 0:
        raise ValueError("ERROR: None of the features from reference were found in the target data")
    if filter_null:
        if exclude_missing is False:
            logger.warning("`filter_null` has not effect because `exclude_missing` is False")
        else:
            t_idx[t_idx] = controlled_compute(
                t_assay.rawData[:, list(t_idx[t_idx].index)][t_assay.cells.active_index('I'), :].sum(axis=0),
                nthreads) != 0
    t_idx = t_idx[t_idx].index
    if exclude_missing:
        s_idx = s_ids.isin(t_ids.values[t_idx])
    else:
        s_idx = s_ids.isin(s_feat_ids)
    s_idx = s_idx[s_idx].index
    t_idx_map = {v: k for k, v in t_ids.to_dict().items()}
    t_re_idx = np.array([t_idx_map[x] if x in t_idx_map else -1 for x in s_ids.values[s_idx]])
    if len(s_idx) != len(t_re_idx):
        raise AssertionError("ERROR: Feature ordering failed. Please report this issue. "
                             f"This is an unexpected scenario. Source has {len(s_idx)} features while target has "
                             f"{len(t_re_idx)} features")
    return s_idx.values, t_re_idx


def align_features(source_assay: Assay, target_assay: Assay, source_cell_key: str,
                   source_feat_key: str, target_feat_key: str, filter_null: bool,
                   exclude_missing: bool, nthreads: int) -> np.ndarray:
    """
    Aligns target features to source features.

    Args:
        source_assay ():
        target_assay ():
        source_cell_key ():
        source_feat_key ():
        target_feat_key ():
        filter_null ():
        exclude_missing ():
        nthreads ():

    Returns:

    """
    from .writers import create_zarr_dataset
    from tqdm import tqdm

    source_feat_ids = source_assay.feats.fetch('ids', key=source_cell_key + '__' + source_feat_key)
    s_idx, t_idx = _order_features(source_assay, target_assay, source_feat_ids, filter_null, exclude_missing, nthreads)
    logger.info(f"{(t_idx == -1).sum()} features missing in target data")
    normed_loc = f"normed__{source_cell_key}__{source_feat_key}"
    norm_params = source_assay.z[normed_loc].attrs['subset_params']
    sorted_t_idx = np.array(sorted(t_idx[t_idx != -1]))

    # TODO: add target cell key
    normed_data = target_assay.normed(target_assay.cells.active_index('I'), sorted_t_idx, **norm_params)
    loc = f"{target_assay.name}/normed__I__{target_feat_key}/data"

    og = create_zarr_dataset(target_assay.z['/'], loc, (1000,), 'float64', (normed_data.shape[0], len(t_idx)))
    pos_start, pos_end = 0, 0
    unsorter_idx = np.argsort(np.argsort(t_idx[t_idx != -1]))
    for i in tqdm(normed_data.blocks, total=normed_data.numblocks[0],
                  desc=f"Writing aligned normed target data to {loc}"):
        pos_end += i.shape[0]
        a = np.ones((i.shape[0], len(t_idx)))
        a[:, np.where(t_idx != -1)[0]] = controlled_compute(i, nthreads)[:, unsorter_idx]
        og[pos_start:pos_end, :] = a
        pos_start = pos_end
    return s_idx



"""
Contains definitions of biological information (e. g. cell cycle genes).
"""
from typing import List

__all__ = ['s_phase_genes', 'g2m_phase_genes']

s_phase_genes: List[str] = [
    'MCM5', 'PCNA', 'TYMS', 'FEN1', 'MCM2', 'MCM4', 'RRM1', 'UNG', 'GINS2', 'MCM6',
    'CDCA7', 'DTL', 'PRIM1', 'UHRF1', 'MLF1IP', 'HELLS', 'RFC2', 'RPA2', 'NASP',
    'RAD51AP1', 'GMNN', 'WDR76', 'SLBP', 'CCNE2', 'UBR7', 'POLD3', 'MSH2', 'ATAD2',
    'RAD51', 'RRM2', 'CDC45', 'CDC6', 'EXO1', 'TIPIN', 'DSCC1', 'BLM', 'CASP8AP2',
    'USP1', 'CLSPN', 'POLA1', 'CHAF1B', 'BRIP1', 'E2F8'
]

g2m_phase_genes: List[str] = [
    'HMGB2', 'CDK1', 'NUSAP1', 'UBE2C', 'BIRC5', 'TPX2', 'TOP2A', 'NDC80', 'CKS2',
    'NUF2', 'CKS1B', 'MKI67', 'TMPO', 'CENPF', 'TACC3', 'FAM64A', 'SMC4', 'CCNB2',
    'CKAP2L', 'CKAP2', 'AURKB', 'BUB1', 'KIF11', 'ANP32E', 'TUBB4B', 'GTSE1',
    'KIF20B', 'HJURP', 'CDCA3', 'HN1', 'CDC20', 'TTK', 'CDC25C', 'KIF2C',
    'RANGAP1', 'NCAPD2', 'DLGAP5', 'CDCA2', 'CDCA8', 'ECT2', 'KIF23', 'HMMR',
    'AURKA', 'PSRC1', 'ANLN', 'LBR', 'CKAP5', 'CENPE', 'CTCF', 'NEK2', 'G2E3',
    'GAS2L3', 'CBX5', 'CENPA'
]
